#include "utils_tx.h"

#include "../createpayload.h"
#include "../elysium.h"
#include "../script.h"
#include "../tx.h"

#include "../../base58.h"
#include "../../coins.h"

#include "../../primitives/transaction.h"

#include "../../script/script.h"
#include "../../script/standard.h"

#include "../../test/test_bitcoin.h"

#include <boost/test/unit_test.hpp>

#include <limits>
#include <vector>

#include <inttypes.h>

namespace elysium {

BOOST_FIXTURE_TEST_SUITE(elysium_sender_firstin_tests, BasicTestingSetup)

/** Creates a dummy class C transaction with the given inputs. */
static CTransaction TxClassC(const std::vector<CTxOut>& txInputs)
{
    CMutableTransaction mutableTx;

    // Inputs:
    for (std::vector<CTxOut>::const_iterator it = txInputs.begin(); it != txInputs.end(); ++it)
    {
        const CTxOut& txOut = *it;

        // Create transaction for input:
        CMutableTransaction inputTx;
        unsigned int nOut = 0;
        inputTx.vout.push_back(txOut);
        CTransaction tx(inputTx);

        // Populate transaction cache:
        ModifyCoin(view, COutPoint(tx.GetHash(), 0),
            [&txOut](Coin & coin){
                coin.out.scriptPubKey = txOut.scriptPubKey;
                coin.out.nValue = txOut.nValue;
            });

        // Add input:
        CTxIn txIn(tx.GetHash(), nOut);
        mutableTx.vin.push_back(txIn);
    }

    // Outputs:
    std::vector<unsigned char> vchPayload = CreatePayload_SimpleSend(1, 1000);
    mutableTx.vout.push_back(EncodeClassC(vchPayload.begin(), vchPayload.end()));

    return CTransaction(mutableTx);
}

/** Helper to create a CTxOut object. */
static CTxOut createTxOut(int64_t amount, const std::string& dest)
{
    return CTxOut(amount, GetScriptForDestination(CBitcoinAddress(dest).Get()));
}

/** Extracts the "first" sender. */
static bool GetFirstSender(const std::vector<CTxOut>& txInputs, std::string& strSender)
{
    int nBlock = std::numeric_limits<int>::max();

    CMPTransaction metaTx;
    CTransaction dummyTx = TxClassC(txInputs);

    if (ParseTransaction(dummyTx, nBlock, 1, metaTx) == 0) {
        strSender = metaTx.getSender();
        return true;
    }

    return false;
}

BOOST_AUTO_TEST_CASE(first_vin_is_sender)
{
    std::vector<CTxOut> vouts;
    vouts.push_back(createTxOut(100, "aByw7PqtCUPj2KggygecNahvPztyFBJw2q")); // Winner
    vouts.push_back(createTxOut(999, "aN5vRJz8YDFUHsffaDqiviifJYvofacfKt"));
    vouts.push_back(createTxOut(200, "a19njnihgJXU4k58KF7phjaLjcMy66d3Mj"));

    std::string strExpected("aByw7PqtCUPj2KggygecNahvPztyFBJw2q");

    std::string strSender;
    BOOST_CHECK(GetFirstSender(vouts, strSender));
    BOOST_CHECK_EQUAL(strExpected, strSender);
}

BOOST_AUTO_TEST_CASE(less_input_restrictions)
{
    std::vector<CTxOut> vouts;
    vouts.push_back(createTxOut(555, "aByw7PqtCUPj2KggygecNahvPztyFBJw2q")); // Winner
    vouts.push_back(PayToPubKey_Unrelated());
    vouts.push_back(PayToBareMultisig_1of3());
    vouts.push_back(NonStandardOutput());

    std::string strExpected("aByw7PqtCUPj2KggygecNahvPztyFBJw2q");

    std::string strSender;
    BOOST_CHECK(GetFirstSender(vouts, strSender));
    BOOST_CHECK_EQUAL(strExpected, strSender);
}

BOOST_AUTO_TEST_CASE(invalid_inputs)
{
    {
        std::vector<CTxOut> vouts;
        vouts.push_back(PayToPubKey_Unrelated());
        std::string strSender;
        BOOST_CHECK(!GetFirstSender(vouts, strSender));
    }
    {
        std::vector<CTxOut> vouts;
        vouts.push_back(PayToBareMultisig_1of3());
        std::string strSender;
        BOOST_CHECK(!GetFirstSender(vouts, strSender));
    }
    {
        std::vector<CTxOut> vouts;
        vouts.push_back(NonStandardOutput());
        std::string strSender;
        BOOST_CHECK(!GetFirstSender(vouts, strSender));
    }
}

BOOST_AUTO_TEST_SUITE_END()

} // namespace elysium
